/*
 * GLX Hardware Device Driver for S3 Savage3D and probably Savage/MX and
 * Savage/IX
 * Copyright (C) 2000 Dominik Behr
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * WITTAWAT YAMWONG, OR ANY OTHER CONTRIBUTORS BE LIABLE FOR ANY CLAIM,
 * DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
 * OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
 * OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 * Based on S3 Virge driver by Jim Duchek <jimduchek@ou.edu>
 *
 * Dominik Behr <behr@promail.pl>
 * thanks to Raja Koduri and Tim Roberts   
 */

#include "context.h"
#include "depth.h"
#include "macros.h"
#include "texstate.h"
#include "triangle.h"
#include "vb.h"
#include "types.h"

#include "glx_log.h"
#include "glx_config.h"
#include "mesaglx/context.h"
#include "mesaglx/matrix.h"
#include "mesaglx/types.h"

#define GC XXGC
#include "gcstruct.h"
#include "pixmapstr.h"
#include "servermd.h" /* PixmapBytePad */
#include "scrnintstr.h"
#include "regionstr.h"
#include "windowstr.h"
#undef GC

#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <sys/mman.h>
#include <stdio.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <signal.h>

#include "glx_symbols.h"
#include "s3savglx.h"

#include "s3savpriv.h"

#include <dlfcn.h>
#include <stdio.h>

/*
 use one surface 4k wide for back and z buffer
*/

extern XSMesaContext XSMesa;

S3SAVDATA stS3Sav;

/* helper function to get symbols from server 
   sets any error if any error happened */
static int any_error;

static void *
my_dlsym(void *handle, const char *name)
{
 void *tmp = dlsym(handle, name);
 char *error;
if ((error = dlerror()) != 0) 
   {
    fputs(error, stderr);
    any_error = 1;
    return 0;
   }
 return tmp;
}

/*******************************************************************************
 hook all symbols we will need from svga server
 symbols should be placed in structure accessed by GLXSYM macro 
*******************************************************************************/
int
s3savHookServerSymbols(void *handle)
{
 unsigned int *ppMmioMem;
 unsigned int *pMem;
 any_error = 0;
 
 ppMmioMem =
 my_dlsym(handle, "s3savMmioMem");
 
 /*
 fprintf(stderr, "mmio mem %08X %08X\n",
         ppMmioMem,
         *ppMmioMem);
  
 pMem = (unsigned int *)(*ppMmioMem);
 
 pMem[0x48500/4] = 0;
 */
 
 stS3Sav.pS3VPriv = my_dlsym(handle, "s3vPriv");
 
 /*
 fprintf(stderr, "s3priv mmiobase %08X\n",stS3Sav.pS3VPriv->MmioBase);
 fprintf(stderr, "s3priv mmiomem %08X\n",stS3Sav.pS3VPriv->MmioMem);
 */
 return !any_error;
}

/*******************************************************************************

*******************************************************************************/
PMemBlock
s3savMakeRoom(unsigned int nSize,
              unsigned int nAlign)
{
 PMemBlock pRes;
 #ifdef DEBUGx
 fprintf(stderr,
         "[s3sav] allocing %d bytes %dKB vidmem\n",
         nSize, nSize / 1024);
 #endif	 

 pRes = mmAllocMem(stS3Sav.pCardHeap,
                   nSize,
                   nAlign,
                   0);
		   
 if (pRes)
    {
     return pRes;
    }
 else
    {
     
     PS3SAVTEXTURE pTex, pPrev;
     if (!stS3Sav.pTextures)
        {
         // awww, crap, texture too big ???
         return 0;
        }
     // find last texture in chain
     pPrev = 0;
     for (pTex = stS3Sav.pTextures; pTex->pNext; pTex = pTex->pNext)
         {
          pPrev = pTex;
         }
     /* now this is tricky, we have to detect if texture is
        still in use or is going to be used by commands in 
	cob */
     /* FIXME
        well, it doesnt work too well
	maybe we should use BCI event command ???
     */	
     #if 0
     if (pTex->nTextureID != 0xFFFFFFFF)
     {
      int i;
      for (i = 0; i < 10; i++)	
          if (((pTex->nTextureID - s3savGetTextureID()) & 0xFFFF) >= 0x8000)
	     break;
          else 
	     usleep(500);
      if (i == 10)
         {
	  stS3Sav.nTextureID = s3savGetTextureID();
	  fprintf(stderr, "[s3sav] something is broken with texture id %08X hardware %08X\n",
	          pTex->nTextureID,
		  s3savGetTextureID());
	 }	     	     
     }
     #endif
     // remove this texture from memory 
     mmFreeMem(pTex->pMemBlock);  
     pTex->pMemBlock = 0;
       
     /* free mesa's link */
     pTex->pMT->DriverData = 0;
     
     // kill the texture
     free((void *)pTex);
     if (pPrev)
        pPrev->pNext = 0;
     else
        stS3Sav.pTextures = 0;   
     return s3savMakeRoom(nSize, nAlign);
    }                  
}

/******************************************************************************
 we do not support any private requests
******************************************************************************/
int 
s3savGLXVendorPrivate(ClientPtr client,
			          XSMesaContext ctx, 
			          xGLXVendorPrivateReq *stuff)
{
 return GLXUnsupportedPrivateRequest;
}

/******************************************************************************
 we dont want to go direct
******************************************************************************/
int 
s3savGLXAllowDirect( ClientPtr client ){
 return 0;
}

/******************************************************************************
 initialize GLX module
******************************************************************************/
GLboolean 
s3savInitGLX( void )
{
 if (!__glx_is_server)
    return GL_FALSE;

 stS3Sav.uVideoMemorySize = GLXSYM(vga256InfoRec).videoRam;

 fprintf(stderr,
         "[s3sav] video memory size %dKB\n",
         stS3Sav.uVideoMemorySize);
         
 stS3Sav.uVideoMemoryPtr = (hwUI32)xf86MapVidMem(0, 
					 LINEAR_REGION, 
					 (pointer)GLXSYM(vga256InfoRec).MemBase,
					 stS3Sav.uVideoMemorySize * 1024);

 fprintf(stderr,
         "[s3sav] video memory ptr %08X\n",
          stS3Sav.uVideoMemoryPtr);
          
 stS3Sav.stFrontBuffer.nWidth = GLXSYM(vga256InfoRec).virtualX;
 stS3Sav.stFrontBuffer.nHeight = GLXSYM(vga256InfoRec).virtualY;
 stS3Sav.stFrontBuffer.nBPP = (GLXSYM(vga256InfoRec).bitsPerPixel + 7) >> 3;
 stS3Sav.stFrontBuffer.nStride = stS3Sav.stFrontBuffer.nWidth * stS3Sav.stFrontBuffer.nBPP;
 stS3Sav.stFrontBuffer.pMemBlock = 0;
 stS3Sav.stFrontBuffer.p = (unsigned char *)stS3Sav.uVideoMemoryPtr;
 
 fprintf(stderr,
         "[s3sav] frontbuffer %dx%dx%d stride %d\n",
         stS3Sav.stFrontBuffer.nWidth,
         stS3Sav.stFrontBuffer.nHeight,
         stS3Sav.stFrontBuffer.nBPP,
         stS3Sav.stFrontBuffer.nStride);
             
 stS3Sav.pCardHeap = mmInit(0, 
                            stS3Sav.uVideoMemorySize * 1024);
    
 /* reserve memory for the front buffer */
 mmReserveMem(stS3Sav.pCardHeap, 
              0, 
              stS3Sav.stFrontBuffer.nStride * stS3Sav.stFrontBuffer.nHeight + 0x1000);

 /* FIXME: what other GLXProcs pointers should we change? */
 GLXProcs.CreateContext = s3savGLXCreateContext;
 GLXProcs.DestroyContext = s3savGLXDestroyContext;
 GLXProcs.SwapBuffers = s3savGLXSwapBuffers;
 GLXProcs.CreateImage = s3savGLXCreateImage;
 GLXProcs.DestroyImage = s3savGLXDestroyImage;
 GLXProcs.CreateDepthBuffer = s3savGLXCreateDepthBuffer;
 GLXProcs.MakeCurrent = s3savGLXMakeCurrent;
 GLXProcs.BindBuffer = s3savGLXBindBuffer;
 GLXProcs.VendorPrivate = s3savGLXVendorPrivate;
 GLXProcs.AllowDirect = s3savGLXAllowDirect;

 stS3Sav.pContexts = 0;
 stS3Sav.pTextures = 0;
 stS3Sav.pContext = 0;

 stS3Sav.pDrawBuffer = 0;
 {
  unsigned int p;
  p = (unsigned int)(stS3Sav.pS3VPriv->MmioMem);
  stS3Sav.pRegs = (PS3SAVREGS)(p + S3SAV_3DREGS);
  stS3Sav.pMMIOMem = (hwUI8 *)p;
  
  stS3Sav.pStatus0 = (hwUI32 *)(p + S3SAV_STATUS0);
  stS3Sav.pStatus1 = (hwUI32 *)(p + S3SAV_STATUS1);
  stS3Sav.pStatus2 = (hwUI32 *)(p + S3SAV_STATUS2);
  fprintf(stderr, "[s3sav] status = %08X\n", *stS3Sav.pStatus0);
 } 
 stS3Sav.nTextureID = s3savGetTextureID();
 
 //stS3Sav.uChangedFlags = 0xFFFFFFFF;
 stS3Sav.stPALADDR = stS3Sav.pRegs->stTEXPALADDR;
 stS3Sav.stTEXADDR = stS3Sav.pRegs->stTEXADDR;
 stS3Sav.stTEXDESC = stS3Sav.pRegs->stTEXDESC;
 stS3Sav.stTEXCTRL = stS3Sav.pRegs->stTEXCTRL;
 stS3Sav.stFOGCTRL = stS3Sav.pRegs->stFOGCTRL;
 stS3Sav.stDRAWCTRL = stS3Sav.pRegs->stDRAWCTRL;
 stS3Sav.stZBCTRL = stS3Sav.pRegs->stZBCTRL;
 stS3Sav.stZBADDR = stS3Sav.pRegs->stZBADDR;
 stS3Sav.stDESTCTRL = stS3Sav.pRegs->stDESTCTRL;
 stS3Sav.stSCSTART = stS3Sav.pRegs->stSCSTART;
 stS3Sav.stSCEND = stS3Sav.pRegs->stSCEND;
 
 {
  hwUI32 nCOB;
  hwUI32 nCOBSize;
  hwUI32 nCOBAddr;
  
  nCOB = S3SAVRD(S3SAV_COB);
  
  nCOBAddr = (nCOB & 0xFFF) << 11;
  nCOBSize = nCOB >> 29;
  switch (nCOBSize) {
   case 0:
    nCOBSize = 256;
    break;
   case 1:
    nCOBSize = 512;
    break;
   case 2:
    nCOBSize = 1024;
    break;        
   case 3:
    nCOBSize = 2048;
    break;    
   case 4:
    nCOBSize = 4096;
    break;    
   case 5:
    nCOBSize = 8192;
    break;    
   case 6:
    nCOBSize = 16384;
    break;    
   case 7:
    nCOBSize = 32768;
    break;    
  }
  fprintf(stderr,
          "[s3sav] COB addr: %08X size: %d\n",
          nCOBAddr,
	  nCOBSize);
  /* reserve memory for Command Overflow Buffer */
  /* add 4KB before COB for cursor */
  mmReserveMem(stS3Sav.pCardHeap, 
               nCOBAddr - 0x1000,
               nCOBSize*4 + 0x1000);
 }

 /* the remaining memory is available for back buffers, depth
    buffers, and textures */
 mmDumpMemInfo(stS3Sav.pCardHeap);

 stS3Sav.pBCI = (hwUI32 *)(stS3Sav.pS3VPriv->BciMem);
 stS3Sav.pfBCI = (float *)(stS3Sav.pS3VPriv->BciMem);
 BCIRST;
 fprintf(stderr, "[s3sav] from mmio[0] = %08X\n",
         *(unsigned int *)(((unsigned int)(stS3Sav.pS3VPriv->MmioMem)) + 0x8000)
        ); 
 #if 0
 {unsigned int uMMIOMem;

  uMMIOMem = (hwUI32)xf86MapVidMem(0, 
				   MMIO_REGION, 
				   (pointer)stS3Sav.pS3VPriv->MmioBase,
			           16*1024*1024);
  fprintf(stderr,
          "[s3sav] mmio mapped at %08X\n", uMMIOMem);

  stS3Sav.pRegs = uMMIOMem + S3SAV_3DREGS;
 }
 
 
 #endif
 {
  unsigned int uRegion;
  for (uRegion = 0; uRegion < 6; uRegion++)
      {
       stS3Sav.aTiledApertures[uRegion] = (hwUI8 *)xf86MapVidMem(0,
                                                                 MMIO_REGION,
								 (pointer)(stS3Sav.pS3VPriv->MmioBase + (uRegion + 1) * 0x1000000),
								 0x1000000);
      }
 }
 fprintf(stderr,
         "[s3sav] 3d regs mapped at %08X size %08X status at %08X\n",
         stS3Sav.pRegs,
         sizeof(S3SAVREGS),
	 stS3Sav.pStatus0);
 fprintf(stderr,
         "[s3sav] BCI mapped at %08X\n",
	 stS3Sav.pBCI);
 fprintf(stderr,
         "[s3sav] s3savInitGLX OK\n");
 return GL_TRUE;
}


/******************************************************************************

******************************************************************************/
void 
s3savGLXCreateDepthBuffer(GLcontext* ctx)
{
 XSMesaContext xsmesa = (XSMesaContext)ctx->DriverCtx;
 PS3SAVBUFFER pBuf; 

 if (xsmesa->xsm_buffer->backimage == NULL) 
    {
     fprintf(stderr,
             "[s3sav] no backimage in xsmesa->sxm_buffer\n");
     return;
    }

 pBuf = (PS3SAVBUFFER)xsmesa->xsm_buffer->backimage->devPriv;


 pBuf->stZBuffer.nBPP = 2; // always do 16 bit z buffer
 // copy size from backbuffer
 pBuf->stZBuffer.nWidth = pBuf->stBackBuffer.nWidth;
 pBuf->stZBuffer.nHeight = pBuf->stBackBuffer.nHeight;
 #ifdef UNIFIED_BUFFER
 if (pBuf->stBackBuffer.pMemBlock)
    {
     pBuf->stZBuffer.nStride = pBuf->stBackBuffer.nStride;
     pBuf->stZBuffer.pMemBlock = pBuf->stBackBuffer.pMemBlock;
     pBuf->stZBuffer.p = stS3Sav.aTiledApertures[APERTURE_ZBUFFER];
     fprintf(stderr, 
             "[s3sav] depth from backbuffer %dx%dx%d stride %d\n",
             pBuf->stZBuffer.nWidth,
	     pBuf->stZBuffer.nHeight,
	     pBuf->stZBuffer.nBPP*8,
	     pBuf->stZBuffer.nStride);
    }
 else
    {
     pBuf->stZBuffer.pMemBlock = 0;
     pBuf->stZBuffer.p = malloc(pBuf->stZBuffer.nHeight 
                                * pBuf->stZBuffer.nStride);
    }    
 #else // not UNIFIED_BUFFER
 pBuf->stZBuffer.nStride = (pBuf->stZBuffer.nWidth * pBuf->stZBuffer.nBPP + 127) & 0xFFFFFF80;
 pBuf->stZBuffer.pMemBlock = s3savMakeRoom(pBuf->stZBuffer.nStride
                                           * ALIGN16(pBuf->stZBuffer.nHeight),
                                           S3SAV_BUFFERALIGN);
        
 if (pBuf->stZBuffer.pMemBlock)
    {
     pBuf->stZBuffer.p = stS3Sav.aTiledApertures[APERTURE_ZBUFFER];
     /*
     pBuf->stZBuffer.p = (unsigned char *)stS3Sav.uVideoMemoryPtr
                       + pBuf->stZBuffer.pMemBlock->ofs;
     */		       
    }
 else
    {
     pBuf->stZBuffer.p = malloc(pBuf->stZBuffer.nHeight 
                                * pBuf->stZBuffer.nStride);
    }
 #endif // UNIFIED_BUFFER
 fprintf(stderr,
         "[s3sav] depth buffer %dx%dx%d at %08X stride %d\n",
         pBuf->stZBuffer.nWidth,
         pBuf->stZBuffer.nHeight,
         pBuf->stZBuffer.nBPP,
         pBuf->stZBuffer.p,
         pBuf->stZBuffer.nStride);
         
 
}


static void
s3savDestroySurface(S3SAVSURFACE *pS)
{
 if (pS->pMemBlock)
    {
     mmFreeMem(pS->pMemBlock);
     pS->pMemBlock = 0;
     pS->p = 0;
     return;
    }
 if (pS->p)
    {
     free(pS->p);
     pS->p = 0;
     return;
    }
}


/******************************************************************************

******************************************************************************/
/*
 * s3savGLXDestroyImage
 */
void 
s3savGLXDestroyImage(GLXImage* image)
{
 PS3SAVBUFFER pBuf;

 pBuf = (PS3SAVBUFFER)image->devPriv;
 if (pBuf)
    {
     #ifdef UNIFIED_BUFFER
     pBuf->stZBuffer.pMemBlock = 0;
     pBuf->stZBuffer.p = 0;
     s3savDestroySurface(&(pBuf->stBackBuffer));
     #else
     s3savDestroySurface(&(pBuf->stZBuffer));
     s3savDestroySurface(&(pBuf->stBackBuffer));
     #endif // UNIFIED_BUFFER
    }
 else 
 if (image->data)
    {
	 free(image->data);
    }

 xfree(image);
}


/******************************************************************************

******************************************************************************/
/*
 * s3savGLXCreateImage
 *
 * This will always succeed, but if the buffer couldn't be allocated in
 * card memory, HW accel won't be usable.
 */
GLXImage *
s3savGLXCreateImage(WindowPtr pwindow, 
                    GLvisual *visual,
                    int width, 
                    int height, 
                    GLXImage *old_image)
{
 PS3SAVBUFFER pBuf;
 GLXImage *pImage;
 int nDepth;
 
 fprintf(stderr, "[s3sav] visual %08X\n",
         visual);
 nDepth = visual->RedBits 
        + visual->GreenBits
        + visual->BlueBits;

 fprintf(stderr,
         "[s3sav] GLXCreateImage %dx%dx%d\n",
         width,
         height,
         nDepth);

 pBuf = (PS3SAVBUFFER)malloc(sizeof(S3SAVBUFFER));
 if (!pBuf)
    {
     FatalError( "Malloc for buffer failed" );
    }
 
 pBuf->stZBuffer.p = 0;
 pBuf->stZBuffer.pMemBlock = 0;
 
 pBuf->stBackBuffer.nWidth = width;
 pBuf->stBackBuffer.nHeight = height;
 pBuf->stBackBuffer.nBPP = (nDepth+7) >> 3;
 #ifdef UNIFIED_BUFFER
 pBuf->stBackBuffer.nStride = 4096;
 #else
 pBuf->stBackBuffer.nStride = (pBuf->stBackBuffer.nWidth * pBuf->stBackBuffer.nBPP + 127) & 0xFFFFFF80; 
 #endif // UNIFIED_BUFFER					 
 fprintf(stderr, "[s3sav] backbuffer %dx%dx%d stride %d\n",
         pBuf->stBackBuffer.nWidth,
	 pBuf->stBackBuffer.nHeight,
	 pBuf->stBackBuffer.nBPP*8,
	 pBuf->stBackBuffer.nStride);
	 
 pBuf->stBackBuffer.pMemBlock = s3savMakeRoom(pBuf->stBackBuffer.nStride *
                                              ALIGN16(pBuf->stBackBuffer.nHeight),
                                              S3SAV_BUFFERALIGN);     
 if (!pBuf->stBackBuffer.pMemBlock)
    {
     // drats, fall back to software
     pBuf->stBackBuffer.p = malloc(pBuf->stBackBuffer.nStride *
                                   pBuf->stBackBuffer.nHeight);
    }
 else
    {
     /*
     pBuf->stBackBuffer.p = (unsigned char *)stS3Sav.uVideoMemoryPtr 
                          + pBuf->stBackBuffer.pMemBlock->ofs;*/
     pBuf->stBackBuffer.p = stS3Sav.aTiledApertures[APERTURE_BACKBUFFER];
    }
	
 pImage = (GLXImage *)xalloc(sizeof(GLXImage));
 if (!pImage) 
    {
     FatalError( "Malloc for back ximage failed" );
    }

 pImage->pwin = pwindow;
 pImage->width = width;
 pImage->height = height;
 pImage->bits_per_pixel = nDepth;
 pImage->data = pBuf->stBackBuffer.p;
 pImage->bytes_per_line = pBuf->stBackBuffer.nStride;
 pImage->devPriv = pBuf;

 pImage->width = pBuf->stBackBuffer.nStride / pBuf->stBackBuffer.nBPP;

 return pImage;
}


/******************************************************************************

******************************************************************************/
/*
 * DoMakeCurrent
 * Changing either the context or the buffer will force this.
 * Redone 3/16/2000 by Justin A. Kolodziej to match what is done in mach64 
 * driver.  
 */
static int 
DoMakeCurrent(XSMesaContext c, 
              XSMesaBuffer b)
{
 PS3SAVCONTEXT pCtx;
 PS3SAVBUFFER pBuf;

 stS3Sav.pContext = 0;
 stS3Sav.pDrawBuffer = 0;

 if (!c || !b) 
    {
     return 0;
    }

 pCtx = (PS3SAVCONTEXT)c->hw_ctx;
 if (!pCtx) 
    {
     FatalError("DoMakeCurrent: for null hardware context");
    }

 /* if we don't have a backimage, create one now.  This happens
  * because of the way software GLX doesn't allocate a
  * buffer when front buffer rendering, but we need one
  * anyway
  */
 if (!b->backimage) 
    {
     GLvisual visual;

     memset( &visual, 0, sizeof( visual ) );
     visual.RedBits = 5;
     visual.GreenBits = 5;
     visual.BlueBits = 5;
     visual.DepthBits = 16;

     b->backimage = s3savGLXCreateImage((WindowPtr)b->frontbuffer,
                                        &visual, 
                                        b->width, 
                                        b->height, 
                                        NULL);
    }

 pBuf = (PS3SAVBUFFER)b->backimage->devPriv;

 s3savSetTile(APERTURE_BACKBUFFER, 
              pBuf->stBackBuffer.pMemBlock->ofs,
	      pBuf->stBackBuffer.nStride,
	      pBuf->stBackBuffer.nBPP);
 if (pBuf->stZBuffer.pMemBlock)
    { 
     s3savSetTile(APERTURE_ZBUFFER, 
                  #ifdef UNIFIED_BUFFER
		  pBuf->stZBuffer.pMemBlock->ofs + HALF_BUFFER,
		  #else
                  pBuf->stZBuffer.pMemBlock->ofs,
		  #endif
                  pBuf->stZBuffer.nStride,
 	          pBuf->stZBuffer.nBPP);
    }

 pCtx->pDrawBuffer = pBuf;
 stS3Sav.pContext = pCtx;
 stS3Sav.pDrawBuffer = pBuf;
 
 #ifdef DEBUGx
 fprintf(stderr, "[s3sav] set drawbuffer %08X\n", stS3Sav.pDrawBuffer);
 #endif
 return 0;
}


/******************************************************************************

******************************************************************************/
/*
 * s3savGLXMakeCurrent
 * Note that GLX calls this every batch of commands,
 * so it will show up several times a frame even if
 * the client application isn't issuing makeCurrent calls.
 */
GLboolean 
s3savGLXMakeCurrent(XSMesaContext c)
{
 if (c == XSMesa) 
    {
     return (GL_TRUE);
    }

 /* mesa 3.1 will sometimes leave unflushed vertexes */
 if (XSMesa) 
    {
     // so why are hacking around here instead of fixing it in mesa ?? TCM
     glFlush();
    }

 XSMesa = c;
 if (c)
    {
	 DoMakeCurrent(c, c->xsm_buffer);
	 gl_make_current(c->gl_ctx,
			         (c->xsm_buffer) ? c->xsm_buffer->gl_buffer : NULL);
    }
 else
    {
     /* detach */
	 DoMakeCurrent( NULL, NULL );
	 gl_make_current( NULL, NULL );
    }

 return GL_TRUE;
}

/******************************************************************************

******************************************************************************/
/*
 * Bind buffer b to context c and make c the current rendering context.
 */
GLboolean 
s3savGLXBindBuffer(XSMesaContext c, 
                   XSMesaBuffer b)
{
 DoMakeCurrent(c, b);
 return XSMesaBindBuffer(c, b);
}


/******************************************************************************

******************************************************************************/
PS3SAVCONTEXT 
s3savCreateContext(GLcontext *ctx)
{
 PS3SAVCONTEXT pCtx;

 pCtx = (PS3SAVCONTEXT)malloc(sizeof(S3SAVCONTEXT));

 if (!pCtx)
    return 0;
    
 pCtx->pDrawBuffer = 0;
 pCtx->gl_ctx = ctx; // set mesa context

 ctx->Const.MaxTextureLevels = 9;
 ctx->Const.MaxTextureUnits = 1;
 ctx->Const.MaxTextureSize = 512;
    
 s3savDDExtensionsInit(ctx);

 return pCtx;
}

/******************************************************************************

******************************************************************************/
/*
 * Create a new XSMesaContext.
 * Input:  v - XSMesaVisual
 *         share_list - another XSMesaContext with which to share display
 *                      lists or NULL if no sharing is wanted.
 * Return:  an XSMesaContext or NULL if error.
 */
XSMesaContext 
s3savGLXCreateContext(XSMesaVisual v,
                      XSMesaContext share_list)
{
 XSMesaContext c;

 c = (XSMesaContext)calloc(1, sizeof(struct xsmesa_context));
    
 if (!c) 
    {
     return NULL;
    }

 c->gl_ctx = gl_create_context(v->gl_visual,
                               (share_list) ? share_list->gl_ctx : NULL,
                               (void *)c,
                               GL_FALSE /* direct rendering */ ); // dont want direct TCM
	   
 if (!c->gl_ctx) 
    {
     free( c );
     return NULL;
    }
        
 c->xsm_visual = v;
 c->xsm_buffer = NULL;  /* set later by XSMesaMakeCurrent */
 c->pixelformat = v->dithered_pf; /* Dithering is enabled by default */
 c->hw_ctx = (void *)s3savCreateContext(c->gl_ctx);
 if (!c->hw_ctx)
    {
     return NULL;
    }
 c->gl_ctx->Driver.UpdateState = s3sav_setup_DD_pointers;

 return c;
}


/******************************************************************************

******************************************************************************/
int 
s3savDestroyContext(PS3SAVCONTEXT pCtx)
{
 if (!pCtx) 
    {
	 return 0;
    }
 s3savWait3DIdle();    
 // FIXME what is the deal with this refcount ?
 // how there may be more than one ref to context ?   
#if 0 /* FIXME: GH */
    s3virgeDestroyContextTextures( ctx );
#endif

 if (pCtx == stS3Sav.pContext) 
    { 
     stS3Sav.pContext = 0;
     stS3Sav.pDrawBuffer = 0;
    }
 free(pCtx);
 return 0;
}


/******************************************************************************

******************************************************************************/
void 
s3savGLXDestroyContext(XSMesaContext c)
{
 if (s3savDestroyContext((PS3SAVCONTEXT)c->hw_ctx ) != 0) 
    {
    }

 XSMesaDestroyContext(c);
 if (XSMesa == c) 
    {
	 XSMesa = 0;
    }
}

